/* CmdBkg (c) 2016-2019 Mikael Sollenborn */

#ifndef UNICODE
#  define UNICODE
#endif
#ifndef _UNICODE
#  define _UNICODE
#endif
#ifdef WINVER
#  undef WINVER
#endif
#define WINVER 0x0501
#ifdef _WIN32_WINNT
#  undef _WIN32_WINNT
#endif
#define _WIN32_WINNT 0x0501

// Usage: cmdbkg "file.bmp" [transparency 0-100(default 33) [includeBorders]]
// Specify no arguments to remove previous background.
// Specify /? as first argument to see this help.

// Compilation: gcc -o cmdbkg.exe cmdbkg.c -lgdi32 -ldwmapi

// TODO/Issues:
// 1. Allow centered placement of image, no stretching?
// 2. Background window visible a short time after restoring minimized console window. Minimize background if console window is minimized?
// 3. Inelegant polling. Use SetWinEventHook instead?
// 4. Find a better way to kill the old background window than this horrible title polling hack

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <dwmapi.h>

#define POLL_INTERVAL 20
#define KILL_TITLE_MESSAGE L" \t \t "

// Functions Fn_LoadBmp and f_SetConsoleTransparency by user "aGerman" at dostips.com

static int Fn_LoadBmp(HWND hWnd, wchar_t *szBmpPath, long x, long y, long z, long w, long h)
{
  HDC hDc = NULL, hDcBmp = NULL;
  HBITMAP hBmp1 = NULL, hBmp2 = NULL;
  HGDIOBJ hGdiObj = NULL;
  BITMAP bmp = { 0 };
  int iRet = EXIT_FAILURE;

  if (hWnd)
  {
    if ((hDc = GetDC(hWnd)))
    {
      if ((hDcBmp = CreateCompatibleDC(hDc)))
      {
        if ((hBmp1 = (HBITMAP)LoadImage(NULL, szBmpPath, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE)))
        {
          if (GetObject(hBmp1, sizeof(bmp), &bmp))
          {
            if (w == -1)
            {
              if ((w = (long)((double)bmp.bmWidth * z / 100.0 + 0.5)) <= 0 || (h = (long)((double)bmp.bmHeight * z / 100.0 + 0.5)) <= 0)
              {
                w = bmp.bmWidth;
                h = bmp.bmHeight;
              }
            }
            if ((hBmp2 = (HBITMAP)CopyImage((HANDLE)hBmp1, IMAGE_BITMAP, w, h, LR_COPYDELETEORG)))
            {
              if ((hGdiObj = SelectObject(hDcBmp, hBmp2)) && hGdiObj != HGDI_ERROR)
              {
                if (BitBlt(hDc, (int)x, (int)y, (int)w, (int)h, hDcBmp, 0, 0, SRCCOPY))
                  iRet = EXIT_SUCCESS;

                DeleteObject(hGdiObj);
              }
              DeleteObject(hBmp2);
            }
          }
          DeleteObject(hBmp1);
        }
        ReleaseDC(hWnd, hDcBmp);
      }
      ReleaseDC(hWnd, hDc);
    }
  }
  return iRet;
}

static BOOL f_SetConsoleTransparency(HWND hConsoleWnd, long percentage)
{
  BYTE bAlpha = 0;
  LONG lNewLong = 0L;
  if (hConsoleWnd && percentage > -1L && percentage < 101L)
  {
    bAlpha = (BYTE)((255L * (100L - percentage) + 50L) / 100L);
    lNewLong = GetWindowLong(hConsoleWnd, GWL_EXSTYLE) | WS_EX_LAYERED;
    if (!SetWindowLong(hConsoleWnd, GWL_EXSTYLE, lNewLong))
      return FALSE;

    return SetLayeredWindowAttributes(hConsoleWnd, 0, bAlpha, LWA_ALPHA);
  }

  return FALSE;
}

int main(void)
{
  const wchar_t scnd[] = L" _secondRun_";
  int bx = 0, by = 0, bw = 0, bh = 0, ret = 0;
  int nbx = 0, nby = 0, nbw = 0, nbh = 0;
  int fw = 0, fh = 0, fth = 0, w = 0, h = 0;
  size_t oldLen = 0;
  HWND hConsoleWnd = NULL, hBkgWnd = NULL, currFgWnd = NULL, tempWnd = NULL;
  RECT bounds = { 0 }, extendBounds = { 0 };
  LPWSTR *szArgList = NULL;
  int argCount = 0, initialRedraw = 10, compTitle = 1;
  long transparency = 33L;
  wchar_t titleBuffer[1024] = { 0 }, *oldCmdLine = NULL, *newCmdLine = NULL, *oldCursor = NULL, *newCursor = NULL, *tmpPtr = NULL;
  WNDCLASS wc = { 0 };
  UINT_PTR timerId = 0;
  BOOL bIsWeirdoBorders = FALSE;
  BOOL res = FALSE;
  MSG msg = { 0 };

  STARTUPINFO si = { 0 };
  PROCESS_INFORMATION pi = { 0 };

  si.cb = sizeof(si);

  oldCmdLine = GetCommandLine();
  szArgList = CommandLineToArgvW(oldCmdLine, &argCount);
  if (szArgList == NULL)
  {
    printf("Cmdbkg error: Unable to get arguments");
    return 1;
  }

  if (argCount == 2 && wcscmp(szArgList[1], L"/?") == 0)
  {
    wprintf(L"\nCmdBkg v1.2 beta : Mikael Sollenborn 2016-2020\n\nUsage: cmdbkg \"file.bmp\" [transparency 0-100(default 33) [includeBorders]]\n\nSpecify no arguments to remove previous background.\n");
    return 0;
  }

  if (argCount < 2 || wcscmp(szArgList[1], L"_secondRun_") != 0)
  {
    oldLen = wcslen(oldCmdLine);
    newCmdLine = malloc(oldLen * sizeof(wchar_t) + sizeof(scnd));
    if (!newCmdLine)
    {
      puts("Cmdbkg error: Unable to allocate memory.");
      return 1;
    }

    oldCursor = wcsstr(oldCmdLine, szArgList[0]) + wcslen(szArgList[0]);
    LocalFree(szArgList);
    tmpPtr = wcspbrk(oldCursor, L" \t");
    oldCursor = tmpPtr ? tmpPtr : oldCmdLine + oldLen;
    newCursor = (wchar_t *)memcpy(newCmdLine, oldCmdLine, (size_t)(oldCursor - oldCmdLine) * sizeof(wchar_t)) + (oldCursor - oldCmdLine);
    newCursor = (wchar_t *)memcpy(newCursor, scnd, sizeof(scnd) - sizeof(wchar_t)) + (sizeof(scnd) - sizeof(wchar_t)) / sizeof(wchar_t);
    memcpy(newCursor, oldCursor, (oldLen - (oldCursor - oldCmdLine) + 1) * sizeof(wchar_t));

    // Start the child process.
    if (!CreateProcess(NULL, // No module name (use command line)
                       newCmdLine, // Command line
                       NULL, // Process handle not inheritable
                       NULL, // Thread handle not inheritable
                       FALSE, // Set handle inheritance to FALSE
                       0, // No creation flags
                       NULL, // Use parent's environment block
                       NULL, // Use parent's starting directory
                       &si, // Pointer to STARTUPINFO structure
                       &pi) // Pointer to PROCESS_INFORMATION structure
    )
    {
      printf("Cmdbkg error: CreateProcess failed (%lu).\n", GetLastError());
      return 1;
    }

    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
    free(newCmdLine);
    return 0;
  }

  hConsoleWnd = GetConsoleWindow();
  if (hConsoleWnd)
  {
    GetWindowRect(hConsoleWnd, &bounds);
    bx = bounds.left;
    by = bounds.top;
    bw = bounds.right - bounds.left;
    bh = bounds.bottom - bounds.top;
  }
  else
  {
    MessageBox(NULL, L"Cmdbkg error: Unable to prepare window creation (1)", L"Error", MB_OK);
    return 1;
  }

  FreeConsole();

  DwmGetWindowAttribute(hConsoleWnd, DWMWA_EXTENDED_FRAME_BOUNDS, &extendBounds, sizeof(extendBounds));
  if (extendBounds.left != bounds.left)
    bIsWeirdoBorders = TRUE;

  f_SetConsoleTransparency(hConsoleWnd, 0);

  GetWindowText(hConsoleWnd, titleBuffer, 1023);
  SetWindowText(hConsoleWnd, KILL_TITLE_MESSAGE);
  Sleep(POLL_INTERVAL * 3);
  SetWindowText(hConsoleWnd, titleBuffer);

  if (argCount < 3)
    return 0;

  if (argCount > 3)
    transparency = wcstol(szArgList[3], NULL, 10);

  if (argCount <= 4 || (argCount > 4 && bIsWeirdoBorders))
  {
    fw = GetSystemMetrics(SM_CXSIZEFRAME);
    if (argCount <= 4)
      fth = GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION);

    fh = fth + GetSystemMetrics(SM_CYSIZEFRAME);
  }

  wc.lpszClassName = L"CmdBkgWindowBackground";
  wc.lpfnWndProc = DefWindowProc; // Use default Window proc
  wc.hInstance = NULL; //GetModuleHandle(0);
  wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  if (!RegisterClass(&wc))
  {
    MessageBox(NULL, L"Cmdbkg error: Unable to prepare window creation (2)", L"Error", MB_OK);
    LocalFree(szArgList);
    return 1;
  }
  hBkgWnd = CreateWindowEx(0, wc.lpszClassName, 0, WS_POPUP | WS_VISIBLE, bx + fw, by + fth, bw - fw * 2, bh - fh, 0, 0, 0, 0);
  if (!hBkgWnd)
  {
    MessageBox(NULL, L"Cmdbkg error: Unable to create background window", L"Error", MB_OK);
    LocalFree(szArgList);
    return 1;
  }

  w = bw - fw * 2, h = bh - fth;
  if (w < 1)
    w = 1;

  if (h < 1)
    h = 1;

  ret = Fn_LoadBmp(hBkgWnd, szArgList[2], 0, 0, 100, w, h);
  if (ret == EXIT_FAILURE)
  {
    MessageBox(NULL, L"Cmdbkg error: Unable to load bitmap", L"Error", MB_OK);
    LocalFree(szArgList);
    return 1;
  }

  f_SetConsoleTransparency(hConsoleWnd, transparency);

  ShowWindow(hBkgWnd, SW_HIDE);
  SetWindowLongPtr(hBkgWnd, GWL_EXSTYLE, WS_EX_TOOLWINDOW);
  ShowWindow(hBkgWnd, SW_SHOW);

  SetForegroundWindow(hConsoleWnd);
  currFgWnd = GetForegroundWindow();

  while ((IsWindow(hConsoleWnd) || IsIconic(hConsoleWnd)) && compTitle != 0)
  {
    timerId = SetTimer(NULL, 0, POLL_INTERVAL, NULL);
    res = GetMessage(&msg, 0, 0, 0);
    KillTimer(NULL, timerId);

    if (!(res && msg.message == WM_TIMER && msg.hwnd == NULL && msg.wParam == timerId))
      DispatchMessage(&msg);

    GetWindowText(hConsoleWnd, titleBuffer, 1023);
    compTitle = wcscmp(titleBuffer, KILL_TITLE_MESSAGE);

    GetWindowRect(hConsoleWnd, &bounds);
    nbx = bounds.left;
    nby = bounds.top;
    nbw = bounds.right - bounds.left;
    nbh = bounds.bottom - bounds.top;

    tempWnd = GetForegroundWindow();
    if (currFgWnd != tempWnd)
    {
      if (initialRedraw < 5)
        initialRedraw = 5;

      currFgWnd = tempWnd;
    }

    LONG cStyle = GetWindowLongA(hConsoleWnd, GWL_STYLE);
    if (cStyle & WS_POPUP)
      SetWindowPos(hBkgWnd, hConsoleWnd, nbx, nby, nbw, nbh, SWP_SHOWWINDOW);
    else
      SetWindowPos(hBkgWnd, hConsoleWnd, nbx + fw, nby + fth, nbw - fw * 2, nbh - fh, SWP_SHOWWINDOW);

    if (initialRedraw > 0 || nbx != bx || nby != by || nbw != bw || nbh != bh)
    {
      if (cStyle & WS_POPUP)
        w = nbw, h = nbh;
      else
        w = nbw - fw * 2, h = nbh - fth;

      if (w < 1)
        w = 1;

      if (h < 1)
        h = 1;

      ret = Fn_LoadBmp(hBkgWnd, szArgList[2], 0, 0, 100, w, h);
      bx = nbx;
      by = nby;
      bw = nbw;
      bh = nbh;
      if (initialRedraw > 0)
        initialRedraw--;
    }
  }

  LocalFree(szArgList);
  return 0;
}
